// DLCLD32.C - fast binary program loader for EVM56K (WIN32)
//-----------------------------------------------------------------------------
// (C) 5.1.1997 Juergen Hasch, hasch@fli.sh.bosch.de
//
//  downloads *.CLD program file to the DSP56002EVM via serial port
//  needs EPROM with CLDLOAD.BIN
//
// The program first reads in the CLD program and analyzes the header.
// After that, all memory blocks are sent to the EVM.
//

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <io.h>
#include <fcntl.h>
#include <conio.h>

// evm56knative.c
int txchar(int);
int rxchar();

// we are on a little endian machine
#define BIG_ENDIAN 0

// include the motorola headers
#include "coreaddr.h"
#include "maout.h"

//------------------------------------------------------------
// define the commands the DSP part of the loader understands:
//

// write data block to P memory
#define SEND_P_BLOCK 1

// write data block to X memory
#define SEND_X_BLOCK 2

// write data block to Y memory
#define SEND_Y_BLOCK 3

// execute the program
#define EXEC_DSP 4
//------------------------------------------------------------
#define DELAY 1000
#define WAITTIME 20UL

struct filehdr fh;
struct aouthdr ah;

int verbose = 0;       // print status messages ?
long start_address;
unsigned char *baseptr;
unsigned char *currnt;

//
//
//
void delay(long d)
{
    long i;
    double x;
    for(i=0; i < d; i++)
        x=(double)d*(double)i;
}


//
// display error message and exit program
//
void error(char *str)
{
 printf("%s\n",str);
 //exit(1);
}

//
// read single character
//
char rd_char()
{
 return *(currnt++);
}

//
// read unsigned long
//
unsigned long rd_ulong()
{
unsigned long val;

  val  = (unsigned long)*(currnt++) << 24;
  val += (unsigned long)*(currnt++) << 16;
  val += (unsigned long)*(currnt++) << 8;
  val += (unsigned long)*(currnt++);

 return val;
}

//
// read unsigned long from given position
//
unsigned long ulong_from(unsigned char *ptr)
{
unsigned long val;

  val  = (unsigned long)*(ptr++) << 24;
  val += (unsigned long)*(ptr++) << 16;
  val += (unsigned long)*(ptr++) << 8;
  val += (unsigned long)*(ptr++);

 return val;
}


//
// read signed long
//
long rd_long()
{
long val;

  val  = ( long)*(currnt++) << 24;
  val += (unsigned long)*(currnt++) << 16;
  val += (unsigned long)*(currnt++) << 8;
  val += (unsigned long)*(currnt++);

 return val;
}

//
// read CORE_ADDR structure
//
void rd_CORE_ADDR(CORE_ADDR *ptr)
{
unsigned long val1,val2;

// lower long word
  ptr->w0.u = rd_ulong();
// upper long word
  ptr->w1.u = rd_ulong();
}

//
// update bytewise CRC calculation
// polynom x^16+x^12+x^5+1
//
unsigned int crcbyte(unsigned int crc,unsigned char data)
{
unsigned int res;
int i;
unsigned long a,b;
unsigned long poly = 0x8408;

 b = data;
 res = crc;

 for (i=0; i<8; i++)
   {
   a =(b & 1) ^ res;

   if (a & 1)
     {
     a = a >> 1;
     a = poly ^ a;
     }
   else
     a = a >> 1;

   b = (b >> 1);
   res = a;
   }
  return res;
}

// ---------------------------------------------------------

//
// read CLD file header
//
void getHdr()
{
  // read file header
  fh.f_magic  = rd_ulong();   // magic number, must be M56KMAGIC
  fh.f_nscns  = rd_long();    // number of sections in CLD file
  fh.f_timdat = rd_long();    // time & date stamp
  fh.f_symptr = rd_long();    // file pointer to start of symbolic table
  fh.f_nsyms  = rd_long();    // number of symbol table entries
  fh.f_opthdr = rd_ulong();   // number of bytes in optional header
  fh.f_flags  = rd_ulong();   // flags

  if (fh.f_magic != M56KMAGIC)
    error("NO DSP56002 CLD file!");

  if (! (fh.f_flags & F_EXEC))
    error("CLD file is not executable!");

  if ( fh.f_opthdr != sizeof(AOUTHDR))
    error("optional runtime header missing!");

  // read optional header
  ah.magic      = rd_long();      // not used
  ah.vstamp     = rd_long();      // not used
  ah.tsize      = rd_long();      // text size in words
  ah.dsize      = rd_long();      // initialized data "  "
  ah.bsize      = rd_long();      // uninitialized data "   "
  rd_CORE_ADDR(&ah.entry); // entry pt.
  rd_CORE_ADDR(&ah.text_start); // base of text used for this file
  rd_CORE_ADDR(&ah.data_start); // base of data used for this file
  rd_CORE_ADDR(&ah.text_end); // end address of text used for this file
  rd_CORE_ADDR(&ah.data_end); // end address of data used for this file

  // get start addresss
  start_address = CORE_ADDR_ADDR(ah.entry);
  if (verbose)
    printf("Start address: %x\n",start_address);
}

//
// read section
//
void getSection()
{
int i;
struct scnhdr sh;
unsigned char *ptr;
char ch,cmd;
unsigned int checksum;
unsigned long rxcheck;

  // get section header
  for (i=0; i < 8; i++)
    sh.s_name[i] = rd_char();
  rd_CORE_ADDR(&sh.s_paddr); // physical address, aliased s_nlib
  rd_CORE_ADDR(&sh.s_vaddr); // virtual address
  sh.s_size    = rd_long();      // section size
  sh.s_scnptr  = rd_long();      // file ptr to raw data for section
  sh.s_relptr  = rd_long();      // file ptr to relocation
  sh.s_lnnoptr = rd_long();      // file ptr to gp histogram
  sh.s_nreloc  = rd_ulong();     // number of relocation entries
  sh.s_nlnno   = rd_ulong();     // number of gp histogram entries
  sh.s_flags   = rd_long();      // flags

  if (verbose)
    printf("Section name: %s\n",sh.s_name);

  // get section
  switch(sh.s_flags)
    {
    case STYP_TEXT:
    case STYP_DATA:
      if ( sh.s_size > 0 && sh.s_scnptr > 0)
	{
	//
	// send data block to DSP card
	//
	ptr = baseptr + sh.s_scnptr;
	//
	// send command for according memory type
	//
	switch(CORE_ADDR_MAP(sh.s_paddr))
	  {
	  case 0:
	    cmd = SEND_P_BLOCK;
	    if (verbose)
	      printf("SEND_P_BLOCK\n");
	    break;
	  case 1:
	    cmd = SEND_X_BLOCK;
	    if (verbose)
	      printf("SEND_X_BLOCK\n");
	    break;
	  case 2:
	    cmd = SEND_Y_BLOCK;
	    if (verbose)
	      printf("SEND_Y_BLOCK\n");
	    break;
	  default:
            if (verbose)
              printf("UNKNOWN TYPE: %d!\n",CORE_ADDR_MAP(sh.s_paddr));
            return; // skip data... (we don't transfer l: memory)
	  }
        delay(2*DELAY);
    txchar(cmd);
    delay(2*DELAY);

	if (verbose)
	  {
	  printf("Address: %04x \n",CORE_ADDR_ADDR(sh.s_paddr));
	  printf("Length: %04x \n",sh.s_size);
	  }

       // check reply
       ch = rxchar();
       delay(2*DELAY);
       if (ch != cmd)
	 printf("incorrect reply: %d (%d)\n",ch,cmd);

    txword(CORE_ADDR_ADDR(sh.s_paddr));        // send address
        delay(2*DELAY);
    txword(sh.s_size);          // send block length

    checksum = 0;
    for (i=0; i < sh.s_size; i++)
	  // send 3 bytes per word
	  {
	  ptr++; // skip high byte
      delay(DELAY);
          checksum=crcbyte(checksum,*(ptr));
      txchar( *(ptr++) );      // send data in bytes
      delay(DELAY);
          checksum=crcbyte(checksum,*(ptr));
      txchar( *(ptr++) );
      delay(DELAY);
          checksum=crcbyte(checksum,*(ptr));
      txchar( *(ptr++) );
	  }
	// send checksum truncated to 16 bit
        delay(2*DELAY);

        rxcheck = 256*(rxchar() & 0xff) + (rxchar() & 0xff);
	if ((checksum&0xffff) != (rxcheck&0xffff))
	  printf("Transmission error!\n");
        if (verbose)
          {
          printf("checksum: %x\n",checksum);
          printf("rxcheck:  %x\n",rxcheck);
          // wait for OK
//          char key=getch();
//          if (key=='x')
//            {
//            delete com;
//            free(baseptr);
//            exit(0);
//            }
          }
        else
          printf(".");
	}
      break;
    }
}

//
// download CLD file
//
int download(char *filename)
{
FILE *fin;                      // file handle
int i,rc;
long temp;
char c,ch;
  long length;
  int  handle;


  //
  // get file length
  //
  handle = open(filename,O_RDONLY);
  length = filelength(handle);
  close(handle);

  // allocate memory for whole CLD file
  if (length > 0)
    baseptr = (unsigned char *)malloc(length);
  else
    error("no cld file!");

  //
  // open file and load complete file into memory
  //
  fin=fopen(filename,"rb");
  if (!fin)
    {
    printf("error opening: %s\n",filename);
    exit(1);
    }

 rc = fread(baseptr, 1, length, fin); // ??
 currnt = baseptr;

 fclose(fin);

 getHdr();
 if (verbose)
   {
   printf("Program entry point: %x\n",CORE_ADDR_ADDR(ah.entry));
   printf("Number of sections: %d\n",fh.f_nscns);
   }

 for (i=0; i < fh.f_nscns; i++)
   {
//   flushInBuf();
   getSection();
   }
  //
  // execute program
  //
  delay(2*DELAY);
  txchar(EXEC_DSP);
  delay(2*DELAY);
  ch = rxchar();
  if (ch != EXEC_DSP)
    printf("ERROR: EXEC_DSP echoed %d (%d)",ch,EXEC_DSP);
  delay(2*DELAY);
 txword(start_address);
  delay(2*DELAY);

  free(baseptr);
}